var scriptName = "nes_emulator";
var scriptAuthor = "Senk Ju";
var scriptVersion = 1.1;

var File = Java.type("java.io.File");
var FileUtils = Java.type("org.apache.commons.io.FileUtils");
var StandardCharsets = Java.type("java.nio.charset.StandardCharsets");
var GL11 = Java.type("org.lwjgl.opengl.GL11");
var ByteBuffer = Java.type("java.nio.ByteBuffer");
var ScaledResolution = Java.type("net.minecraft.client.gui.ScaledResolution");
var Gui = Java.type("net.minecraft.client.gui.Gui");
var Keyboard = Java.type("org.lwjgl.input.Keyboard");
var GlStateManager = Java.type("net.minecraft.client.renderer.GlStateManager");
var Paths = Java.type("java.nio.file.Paths");
var Files = Java.type("java.nio.file.Files");
var System = Java.type("java.lang.System");
var Thread = Java.type("java.lang.Thread");

var lock = {};

script.import("nes_emulator_files/nes.js");
script.import("nes_emulator_files/controller.js");

function log(message, error) {
    chat.print("§8[§9NESEmulator§8] " + (error ? "§c" : "§f") + message);
}

function readRom(fileName) {
    var file = new File(mc.mcDataDir + "/LiquidBounce-1.8/scripts/nes_roms/" + fileName);
    return FileUtils.readFileToString(file, StandardCharsets.ISO_8859_1);
}

function getInstalledRoms() {
    var romsPath = Files.walk(Paths.get(mc.mcDataDir + "/LiquidBounce-1.8/scripts/nes_roms"));
    var roms = [];

    romsPath.filter(function (file) {
        return file.toString().endsWith(".nes");
    }).map(function (file) {
        return file.toString().split(/\\|\//g).pop();
    }).forEach(function (file) {
        roms.push(file);
    });

    return roms;
}

function NesEmulatorModule() {

    var nes = null;
    var currentFrame = null;
    var width = 256;
    var height = 240;
    var enabled = false;

    var texture = null;

    var installedRoms = getInstalledRoms();
    var romValue = value.createList("ROM", installedRoms, installedRoms[0]);

    this.getName = function () {
        return "NESEmulator";
    };

    this.getDescription = function () {
        return "Allows you to play NES games in LiquidBounce.";
    };

    this.getCategory = function () {
        return "Misc";
    };

    this.addValues = function (values) {
        values.add(romValue);
    };

    this.setButtonPressed = function (key, pressed) {
        if (pressed) {
            nes.buttonDown(1, key);
        } else {
            nes.buttonUp(1, key);
        }
    };

    this.onEnable = function () {
        log("Loading emulator...", false);

        enabled = true;

        nes = new NES({
            onFrame: function (frameBuffer) {
                Java.synchronized(function () {
                    currentFrame = ByteBuffer.allocateDirect(frameBuffer.length * 3);
                    for (var i = 0; i < frameBuffer.length; i++) {
                        var pixel = frameBuffer[i];

                        currentFrame.put(pixel & 0xFF);
                        currentFrame.put((pixel >> 8) & 0xFF);
                        currentFrame.put((pixel >> 16) & 0xFF);
                    }
                    currentFrame.flip();
                }, lock)();
            }
        });

        texture = GL11.glGenTextures();

        GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture);

        GL11.glPixelStorei(GL11.GL_UNPACK_ALIGNMENT, 1);

        GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MIN_FILTER, GL11.GL_NEAREST);
        GL11.glTexParameteri(GL11.GL_TEXTURE_2D, GL11.GL_TEXTURE_MAG_FILTER, GL11.GL_NEAREST);

        GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);

        var thread = new Thread(function () {

            try {
                nes.loadROM(readRom(romValue.get()));
            } catch (err) {
                log("Failed to laod rom " + romValue.get(), true);
            }

            var lastLoopTime;
            var TARGET_FPS = 60;
            var OPTIMAL_TIME = 1000000000 / TARGET_FPS;

            while (enabled) {
                lastLoopTime = System.nanoTime();

                nes.frame();

                var time = (lastLoopTime - System.nanoTime() + OPTIMAL_TIME) / 1000000;
                if (time >= 0) {
                    Thread.sleep(time);
                }
            }
        });

        thread.start();

        log("Performance will improve over time!", false);
    };

    this.onDisable = function () {
        enabled = false;

        if (texture !== null) {
            GL11.glDeleteTextures(texture);
        }
    }

    this.onUpdate = function () {
        if (nes === null) {
            return;
        }

        this.setButtonPressed(Controller.BUTTON_A, Keyboard.isKeyDown(Keyboard.KEY_A));
        this.setButtonPressed(Controller.BUTTON_B, Keyboard.isKeyDown(Keyboard.KEY_B));
        this.setButtonPressed(Controller.BUTTON_SELECT, Keyboard.isKeyDown(Keyboard.KEY_RETURN));
        this.setButtonPressed(Controller.BUTTON_START, Keyboard.isKeyDown(Keyboard.KEY_SPACE));
        this.setButtonPressed(Controller.BUTTON_UP, Keyboard.isKeyDown(Keyboard.KEY_UP));
        this.setButtonPressed(Controller.BUTTON_DOWN, Keyboard.isKeyDown(Keyboard.KEY_DOWN));
        this.setButtonPressed(Controller.BUTTON_LEFT, Keyboard.isKeyDown(Keyboard.KEY_LEFT));
        this.setButtonPressed(Controller.BUTTON_RIGHT, Keyboard.isKeyDown(Keyboard.KEY_RIGHT));
    };

    this.onRender2D = function () {
        if (nes === null) {
            return;
        }

        if (currentFrame === null) {
            return;
        }

        var resolution = new ScaledResolution(mc);

        var x = resolution.getScaledWidth() / 2 - width / 2;
        var y = resolution.getScaledHeight() / 2 - height / 2;

        GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture);

        Java.synchronized(function () {
            GL11.glTexImage2D(GL11.GL_TEXTURE_2D, 0, GL11.GL_RGB, width, height, 0, GL11.GL_RGB, GL11.GL_UNSIGNED_BYTE, currentFrame);
        }, lock)();

        GL11.glBindTexture(GL11.GL_TEXTURE_2D, 0);

        GL11.glBindTexture(GL11.GL_TEXTURE_2D, texture);
        Gui.drawScaledCustomSizeModalRect(x, y, 0, 0, width, height, width, height, width, height);
    };
}

var nesEmulatorModule = new NesEmulatorModule();

function onEnable() {
    moduleManager.registerModule(nesEmulatorModule);
}

function onDisable() {
    moduleManager.unregisterModule(nesEmulatorModule);
}